/* GPIO for a mechanical rotary encoder with push button, a push button and a reed
   contact
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "driver/gpio.h"
#include "esp_sleep.h"

#include "soc/rtc_cntl_reg.h"
#include "soc/sens_reg.h"
#include "soc/rtc_periph.h"
#include "driver/rtc_io.h"
#include "driver/rtc_cntl.h"
#include "esp32/ulp.h"
#include "ulp_main.h"

/*
   RTC_GPIOs used:
   Name                 RTC_IO num    GPIO num         sleep mask
   Reed contact         RTC_GPIO00    GPIO36           x10
   Rotary B             RTC_GPIO03    GPIO39           x80
   Rotary A             RTC_GPIO04    GPIO34           x04
   Rotary push          RTC_GPIO05    GPIO35           x08
   Push 1               RTC_GPIO08    GPIO33           x02

   GPIO 34-39 are input only and do not feature pull-up / pull-down
   Internal pull-up ( pull-down are not used to conserve power
 */

extern const uint8_t ulp_main_bin_start[] asm("_binary_ulp_main_bin_start");
extern const uint8_t ulp_main_bin_end[]   asm("_binary_ulp_main_bin_end");

static void IRAM_ATTR ulp_isr(void *);
static void init_ulp_program(void);

static int volatile cnt_int;
static int volatile cnt_prs;
static int volatile cnt_rls;
static int volatile cnt_rot;

static int volatile cnt_reed;

void app_main(void)
{
    int val_btn = 0;

    cnt_rot = 0;
    cnt_prs = 0;
    cnt_rls = 0;
    cnt_reed = 0;

    /* init ULP and RTC GPIO */
    printf("Not ULP wakeup, initializing ULP\n");
    init_ulp_program();
    vTaskDelay(1 / portTICK_RATE_MS);  /* need to wait for uLP startup*/

    /* register ULP interrupt */
    printf("register ULP isr ");
    ESP_ERROR_CHECK(rtc_isr_register(ulp_isr, NULL, RTC_CNTL_SAR_INT_ST_M));

    /* enableULP interrupt */
    REG_SET_BIT(RTC_CNTL_INT_ENA_REG, RTC_CNTL_ULP_CP_INT_ENA_M);

    for(;;){
    vTaskDelay(1);
    printf("cnt_reed, cnt_prs, cnt_rls, cnt_rot: %4x, %4x, %4x, %6d\n", cnt_reed, cnt_prs, cnt_rls, cnt_rot);
    }

    /* read all inputs from above*/
    val_btn = rtc_gpio_get_level(GPIO_NUM_36);
    val_btn <<= 1;
    val_btn |= rtc_gpio_get_level(GPIO_NUM_33);
    val_btn <<= 1;
    val_btn |= rtc_gpio_get_level(GPIO_NUM_35);
    val_btn <<= 1;
    val_btn |= rtc_gpio_get_level(GPIO_NUM_34);
    val_btn <<= 1;
    val_btn |= rtc_gpio_get_level(GPIO_NUM_39);

    printf("Inputs read\n");
    printf("val_bin, cnt_rot: %02x, %02x\n", val_btn, cnt_rot);

    vTaskDelay(5000 / portTICK_RATE_MS);

    printf("Hibernate\n");
    esp_sleep_pd_config(ESP_PD_DOMAIN_RTC_PERIPH,   ESP_PD_OPTION_OFF);
    esp_sleep_pd_config(ESP_PD_DOMAIN_RTC_SLOW_MEM, ESP_PD_OPTION_OFF);
    esp_sleep_pd_config(ESP_PD_DOMAIN_RTC_FAST_MEM, ESP_PD_OPTION_OFF);
    esp_sleep_pd_config(ESP_PD_DOMAIN_XTAL,         ESP_PD_OPTION_OFF);
    esp_deep_sleep_start();

}


static void IRAM_ATTR ulp_isr(void* arg)
/**
 * ULP interrupt service routine. ULP interrupt triggered when ULP executes 'wake' instruction
 * and the Xtensa is not sleeping.
 */
{
 // disable interrupt; disable is needed as ULP int is triggered by a (150kHz) slow clock
 /* REG_CLR_BIT(RTC_CNTL_INT_ENA_REG, RTC_CNTL_ULP_CP_INT_ENA_M);
 ets_delay_us(20); */

 cnt_int++;

 if ((ulp_evnt & 0x07) == 1){
  /* button pressed */
  // rtc_gpio_set_level(GPIO_NUM_2, 1);
  cnt_prs++;
  if (ulp_btn & 0x0001)
    cnt_reed++;
  }
 if ((ulp_evnt & 0x07) == 2){
  /* button released */
  // rtc_gpio_set_level(GPIO_NUM_2, 0);
  cnt_rls++;
  }

 cnt_rot += (short int) ulp_rot_st;

 /* clear event and interrupt */
 ulp_evnt = 0;
 ulp_rot_st = 0;
 WRITE_PERI_REG(RTC_CNTL_INT_CLR_REG, READ_PERI_REG(RTC_CNTL_INT_ST_REG));

 /* re-enable ULP interrupt */
 /* REG_SET_BIT(RTC_CNTL_INT_ENA_REG, RTC_CNTL_ULP_CP_INT_ENA_M); */

}

static void init_ulp_program(void)
{
    esp_err_t err = ulp_load_binary(0, ulp_main_bin_start,
            (ulp_main_bin_end - ulp_main_bin_start) / sizeof(uint32_t));
    ESP_ERROR_CHECK(err);

    /* Initialize some variables used by ULP program.
     * Each 'ulp_xyz' variable corresponds to 'xyz' variable in the ULP program.
     * These variables are declared in an auto generated header file,
     * 'ulp_main.h', name of this file is defined in component.mk as ULP_APP_NAME.
     * These variables are located in RTC_SLOW_MEM and can be accessed both by the
     * ULP and the main CPUs.
     *
     * Note that the ULP reads only the lower 16 bits of these variables.
     */
    ulp_btn = 0;
    ulp_evnt = 0;

    /* Initialize selected GPIO as RTC IO, enable input, disable pullup and pulldown */
    rtc_gpio_init(GPIO_NUM_36);                     // after sleep io mode is RTC, but to be sure...
    rtc_gpio_set_direction(GPIO_NUM_36, RTC_GPIO_MODE_INPUT_ONLY); // direction input (also needed for input only pins)
    rtc_gpio_pulldown_dis(GPIO_NUM_36);
    rtc_gpio_pullup_dis(GPIO_NUM_36);

    rtc_gpio_init(GPIO_NUM_39);
    rtc_gpio_set_direction(GPIO_NUM_39, RTC_GPIO_MODE_INPUT_ONLY);
    rtc_gpio_pulldown_dis(GPIO_NUM_39);
    rtc_gpio_pullup_dis(GPIO_NUM_39);

    rtc_gpio_init(GPIO_NUM_34);
    rtc_gpio_set_direction(GPIO_NUM_34, RTC_GPIO_MODE_INPUT_ONLY);
    rtc_gpio_pulldown_dis(GPIO_NUM_34);
    rtc_gpio_pullup_dis(GPIO_NUM_34);

    rtc_gpio_init(GPIO_NUM_35);
    rtc_gpio_set_direction(GPIO_NUM_35, RTC_GPIO_MODE_INPUT_ONLY);
    rtc_gpio_pulldown_dis(GPIO_NUM_35);
    rtc_gpio_pullup_dis(GPIO_NUM_35);

    rtc_gpio_init(GPIO_NUM_33);
    rtc_gpio_set_direction(GPIO_NUM_33, RTC_GPIO_MODE_INPUT_ONLY);
    rtc_gpio_pulldown_dis(GPIO_NUM_33);
    rtc_gpio_pullup_dis(GPIO_NUM_33);

    rtc_gpio_init(GPIO_NUM_2);
    rtc_gpio_set_direction(GPIO_NUM_2, RTC_GPIO_MODE_OUTPUT_ONLY);
    rtc_gpio_pulldown_dis(GPIO_NUM_2);
    rtc_gpio_pullup_dis(GPIO_NUM_2);


    /* Disconnect GPIO12 and GPIO15 to remove current drain through
     * pullup/pulldown resistors.
     * GPIO12 may be pulled high to select flash voltage.
     */
    rtc_gpio_isolate(GPIO_NUM_12);
    rtc_gpio_isolate(GPIO_NUM_15);

    /* Mask for wake-up by means of ext1 */
    esp_sleep_enable_ext1_wakeup(0x9e00000000, ESP_EXT1_WAKEUP_ANY_HIGH);

    printf("RTC GPIO initialized\n");

    //esp_deep_sleep_disable_rom_logging(); // suppress boot messages

    /* Start the program */
    err = ulp_run(&ulp_entry - RTC_SLOW_MEM);
    ESP_ERROR_CHECK(err);
}
