/* ULP assembly files are passed through C preprocessor first, so include directives
   and C macros may be used in these files
*/

#include "soc/rtc_cntl_reg.h"
#include "soc/rtc_io_reg.h"
#include "soc/soc_ulp.h"

/*
   RTC_GPIOs used:
   Name                 RTC_IO num    GPIO num         sleep mask
   Reed contact         RTC_GPIO00    GPIO36           x10
   Rotary B             RTC_GPIO03    GPIO39           x80
   Rotary A             RTC_GPIO04    GPIO34           x04
   Rotary push          RTC_GPIO05    GPIO35           x08
   Push 1               RTC_GPIO08    GPIO33           x02

   GPIO 34-39 are input only and do not feature pull-up / pull-down
   Internal pull-up ( pull-down are not used to conserve power

   ULP register usage:
   r0   temp
   r1 current value
   r2 previous value
   r4 memory address

*/

  /* Define variables, which go into .bss section (zero-initialized data) */
  .bss
  /* mirrors gpio of buttons and encoder AB */
  .global btn
btn:
  .long 0

  .global prev_io
prev_io:
  .long 0

  .global curr_io
curr_io:
  .long 0

  .global rot_ab
rot_ab:
  .long 0

  .global rot_st
rot_st:
  .long 0

  /* Event 1 = btn pressed, 2 = brn released, 4 = rotated */
  .global evnt
evnt:
  .long 0

  /* Code goes into .text section */
  .text

  .global st_tbl
st_tbl:
  .long 0
  .long -1
  .long 1
  .long 0
  .long 1
  .long 0
  .long 0
  .long -1
  .long -1
  .long 0
  .long 0
  .long 1
  .long 0
  .long 1
  .long -1
  .long 0

  .global entry

entry:
  /* establish initial values */
  READ_RTC_REG(RTC_GPIO_IN_REG, RTC_GPIO_IN_NEXT_S, 9)
  move r2, r0
  and r0, r0, 0x0139
  move r3, prev_io
  st r0, r3, 0

  and r0, r0, 0x0018
  rsh r0, r0, 0x03
  move r3, rot_ab
  st r0, r3, 0

loop_forever:

  /* Read the value of lower 9 RTC IOs into R0 */
  READ_RTC_REG(RTC_GPIO_IN_REG, RTC_GPIO_IN_NEXT_S, 9)

  /* mask unused ios; 0x0139 = 313 */
  and r0, r0, 0x0139

  /* current value to r1 */
  move r1, r0

  /* store in variable io */
  move r3, curr_io
  st r1, r3, 0

  sub r0, r1, r2

  /* no io changed */
  jump loop_forever, EQ

  /* wait for bouncing to stop */
  /* WAIT 65535 equals approx. 8.19ms 1 wait cyvle = 1.0 / 1e6 */
  WAIT 32768

  /* Read again the value of lower 9 RTC IOs into R0 */
  READ_RTC_REG(RTC_GPIO_IN_REG, RTC_GPIO_IN_NEXT_S, 9)

  /* mask unused io; 0x0139 = 313 */
  and r0, r0, 0x0139
  sub r0, r0, r1

  /* io changed */
  jump io_chg, EQ
  jump loop_forever

io_chg:
  and r1, r1, 0x0018
  and r2, r2, 0x0018

  sub r0, r1, r2
  jump btn_chg, EQ

  /* encoder AB changed */
  move r3, evnt
  move r0, 4
  st r0, r3, 0

  /* decode AB */
  /* WRITE_RTC_REG(RTC_GPIO_OUT_REG,RTC_GPIO_OUT_DATA_S+12,1,1) */
  move r3, curr_io
  ld r1, r3, 0
  rsh r1, r1, 3

  move r3, rot_ab
  ld r0, r3, 0
  lsh r0, r0, 2
  or r0, r0, r1
  and r0, r0, 0x000f

  /* table look-up */
  add r1, r0, st_tbl
  ld r0, r1, 0

  /* store in rot_st */
  move r3, rot_st
  st r0, r3, 0

  /* store current ab as previous ab */
  move r3, curr_io
  ld r0, r3, 0
  rsh r0, r0, 3
  and r0, r0, 0x0003

  move r3, rot_ab
  st r0, r3, 0

  /* restore r1 and r2 */
  move r3, curr_io
  ld r1, r3, 0
  move r2, r1

  /* trigger ULP int to notify Xtensa */
  wake

  jump loop_forever

btn_chg:
  WRITE_RTC_REG(RTC_GPIO_OUT_REG,RTC_GPIO_OUT_DATA_S+12,1,0)
  move r3, curr_io
  ld r1, r3, 0
  and r1, r1, 0x0121
  move r3, prev_io
  ld r2, r3, 0
  and r2, r2, 0x0121

  sub r0, r1, r2
  jump h_to_l, OV

  /* WRITE_RTC_REG(RTC_GPIO_OUT_REG,RTC_GPIO_OUT_DATA_S+12,1,1) */
  move r3, evnt
  move r0, 1
  st r0, r3, 0
  jump cont_1

h_to_l:
  /* WRITE_RTC_REG(RTC_GPIO_OUT_REG,RTC_GPIO_OUT_DATA_S+12,1,0) */
  move r3, evnt
  move r0, 2
  st r0, r3, 0

cont_1:
  /* store to variable btn */
  move r3, btn
  st r1, r3, 0

  /* restore r1 */
  move r3, curr_io
  ld r1, r3, 0

  /* previous value = current value */
  move r2, r1
  move r3, prev_io
  st r2, r3, 0

  /* trigger ULP int to notify Xtensa */
  wake

  jump loop_forever
